# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""BasePF9SAMLPlugin class."""

import base64
import urlparse

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.identity import v3


class _Pf9Saml2TokenAuthMethod(v3.AuthMethod):
    _method_parameters = []  # type: List[str]

    def get_auth_data(self, session, auth, headers, **kwargs):
        raise exceptions.MethodNotImplemented('This method should never '
                                              'be called')


class BasePF9SAMLPlugin(v3.FederationBaseAuth):
    """Base PF9 SAML authentication class.

    IDP specific classes should be based off of this.
    """

    _auth_method_class = _Pf9Saml2TokenAuthMethod

    def __init__(
        self,
        auth_url,
        username,
        password,
        protocol,
        identity_provider,
        **kwargs
    ):
        """
        Class constructor accepting following parameters.

        :param auth_url: URL of the Identity Service
        :type auth_url: string
        :param username: User's login
        :type username: string
        :param password: User's password
        :type password: string
        :param protocol: Protocol to be used for the authentication.
                         The name must be equal to one configured at the
                         keystone sp side. This value is used for building
                         dynamic authentication URL.
                         Typical value would be: saml2
        :type protocol: string
        :param identity_provider: Name of the Identity Provider the client
                                  will authenticate against. This parameter
                                  will be used to build a dynamic URL used to
                                  obtain unscoped OpenStack token.
        :type identity_provider: string
        """
        super(BasePF9SAMLPlugin, self).__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            **kwargs)
        self.username = username
        self.password = password
        self.__redirect_url = None

        _auth_url = urlparse.urlparse(auth_url)
        self._pf9_endpoint = "{0}://{1}".format(
            _auth_url.scheme,
            _auth_url.netloc,
        )

    def _authenticate(self, session):
        """Authenticate with identity provider.

        :param session
        :type session: keystoneauth1.session.Session
        """
        raise NotImplementedError

    def _cookies(self, session):
        """
        Check if cookie jar is not empty.

        keystoneauth1.session.Session object doesn't have a cookies attribute.
        We should then try fetching cookies from the underlying
        requests.Session object. If that fails too, there is something wrong
        and let Python raise the AttributeError.
        :param session
        :type session: keystoneauth1.session.Session
        :returns: Boolean indicating MFA support
        :rtype: True if cookie jar is nonempty, False otherwise
        :raises AttributeError: in case cookies are not find anywhere
        """
        try:
            return bool(session.cookies)
        except AttributeError:
            pass

        return bool(session.session.cookies)

    def _get_auth_cookie(self, session, saml_assert):
        """Provide authenticated SAML assertion & obtain unscoped token.

        :param session
        :type session: keystoneauth1.session.Session
        :param saml_assert: SAML assertion
        :type saml_assert: str
        """
        url = self._pf9_endpoint + '/Shibboleth.sso/SAML2/POST'
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        payload = {'SAMLResponse': base64.b64encode(saml_assert)}

        self._post_call(session, url, headers, payload)

    def _get_redirect_url(self):
        try:
            response = self.session.get(
                self._pf9_endpoint + "/Shibboleth.sso/Login",
                authenticated=False,
                redirect=False)

        except exceptions.http.InternalServerError as excp:
            raise exceptions.InternalServerError(
                http_status=excp.http_status,
                message="Unable to determine Identity Provider redirect URL"
                " from Service Provivder.",
                details=excp.message,
                method=excp.method,
                url=excp.url)

        if response.status_code == 302:
            self.__redirect_url = response.headers["Location"]

            return self.__redirect_url
        else:
            return None

    def _mfa_supported(self):
        """Check if MFA is supported."""
        raise NotImplementedError

    def _post_call(self, session, url, headers, payload, authenticated=False):
        """Helper method to POST data with the correct content type.

        :param session
        :type session: keystoneauth1.session.Session
        :param url: Target URL
        :param headers: HTTP request headers
        :param payload: Request payload
        :type url: str
        :type headers: dict
        :type payload: str
        """
        post_args = dict(headers=headers)
        if headers['Content-Type'] == 'application/x-www-form-urlencoded':
            post_args['data'] = payload
        elif headers['Content-Type'] == 'application/json':
            post_args['json'] = payload
        return session.post(
            url,
            authenticated=authenticated,
            redirect=False,
            **post_args
        )

    def _redirect_url(self):
        """Return redirect url."""
        return self.__redirect_url or self._get_redirect_url()

    def get_unscoped_auth_ref(self, session, **kwargs):
        """Obtain unscoped token after authenticating with SAML IdP.

        :param session:
        :type session: keystoneauth1.session.Session
        """
        try:
            saml_response = self._authenticate(session)
        except Exception as e:
            raise e

        # Exit if authentication failed
        if saml_response is False:
            raise exceptions.AuthorizationFailure(
                "Unable to obtain SAML assertion. Authentication failed or"
                " server error.")

        self._get_auth_cookie(session, saml_response)

        if self._cookies(session) is False:
            raise exceptions.AuthorizationFailure(
                "Session object doesn't contain a cookie, therefore you are "
                "not allowed to enter the Service Provider's protected area.")

        resp = session.get(self.federated_token_url,
                           authenticated=False,
                           cookies=session.session.cookies.get_dict())

        return access.create(body=resp.json(), resp=resp)



# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Okta SAML authentication driver."""
import re
import urlparse
from oktaauth import models
from pf9_saml_auth.v3 import base


class Password(base.BasePF9SAMLPlugin):
    """okta.Password class."""

    def __init__(self, **kwargs):
        """Class constructor accepting following parameters.

        Inherits input parameters from parent class.
        """
        self.session = None

        super(Password, self).__init__(**kwargs)

        if not self._mfa_supported():
            self._mfa_passcode = None

    def _app_info(self):
        """Okta-specific application info retrieved from the URL.

        :returns: Tuple containing app type & ID
        :rtype: tuple
        """
        redirect_url = urlparse.urlparse(self._redirect_url())
        if re.search("okta", redirect_url.hostname):
            app_info = re.match(
                r"^\/app\/(\w+)\/(\w+)\/sso/saml$",
                redirect_url.path
            )
            return app_info.groups(0)

    def _authenticate(self, session):
        """Authenticate with identity provider.

        :param session
        :type session: keystoneauth1.session.Session
        :returns: SAML response
        :rtype: str
        """
        if self.session is None:
            self.session = session

        app_type, app_id = self._app_info()
        okta = models.OktaSamlAuth(
            urlparse.urlparse(self._redirect_url()).hostname,
            app_type,
            app_id,
            self.username,
            self.password,
            self._mfa_passcode,
        )

        try:
            saml_response = okta.auth()

            return saml_response
        except Exception as excp:
            raise excp

    def _mfa_supported(self):
        """Check if MFA is supported.

        :returns: Boolean indicating if MFA is supported.
        :rtype: True if MFA supported, False otherwise
        """
        return False



# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Platform9 ADFS authentication driver."""
from keystoneauth1.extras._saml2 import V3ADFSPassword
from keystoneauth1 import access


def __pf9_adfs_init(self, service_provider_endpoint,
                    service_provider_entity_id=None, **kwargs):
    """Constructor for ``ADFSPassword``.

    :param service_provider_endpoint: Endpoint where an assertion is being
            sent, for instance: ``https://host.domain/Shibboleth.sso/ADFS``
    :type service_provider_endpoint: string
    :param service_provider_entity_id: SP SAML entity ID
    :type service_provider_entity_id: string

    Remaining arguments inherit from parent class.
    """
    super(V3ADFSPassword, self).__init__(**kwargs)

    self.service_provider_endpoint = service_provider_endpoint
    self.service_provider_entity_id = service_provider_entity_id \
        or service_provider_endpoint


def __pf9_set_wsa_address(self, address):
    """Set WS-Trust Address.

    :param address:
    :type address: str
    """
    APPLIES_TO_NAMESPACES = self.NAMESPACES
    APPLIES_TO_NAMESPACES.update({
        'trust': 'http://docs.oasis-open.org/ws-sx/ws-trust/200512',
        'wsp': 'http://schemas.xmlsoap.org/ws/2004/09/policy',
        'wsa': 'http://www.w3.org/2005/08/addressing'

    })

    wsa_address_xpath = ''.join((
        '/s:Envelope/s:Body/trust:RequestSecurityToken/wsp:AppliesTo/',
        'wsa:EndpointReference/wsa:Address'
    ))

    # Update WSA address
    self.prepared_request.xpath(
        wsa_address_xpath,
        namespaces=APPLIES_TO_NAMESPACES
    )[0].text = address


def __pf9_get_unscoped_auth_ref(self, session, *kwargs):
    """Obtain unscoped token after authenticating with SAML IdP.

    Replace V3ADFSPassword.get_unscoped_auth_ref() to add
    __pf9_set_wsa_address() function.

    :param session:
    :type session: keystoneauth1.session.Session
    """
    self._prepare_adfs_request()
    self.__pf9_set_wsa_address(self.service_provider_entity_id)
    self._get_adfs_security_token(session)
    self._prepare_sp_request()
    self._send_assertion_to_service_provider(session)
    self._access_service_provider(session)

    return access.create(resp=self.authenticated_response)


# Monkey patch V3ADFSPassword with Platform9 changes.
V3ADFSPassword.__init__ = __pf9_adfs_init
V3ADFSPassword.__pf9_set_wsa_address = __pf9_set_wsa_address
V3ADFSPassword.get_unscoped_auth_ref = __pf9_get_unscoped_auth_ref



# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from pf9_saml_auth import v3

_ADFS_AVAILABLE = v3._ADFS_AVAILABLE
_OKTA_AVAILABLE = v3._OKTA_AVAILABLE
_ONELOGIN_AVAILABLE = v3._ONELOGIN_AVAILABLE

V3Pf9ADFSPassword = v3.ADFSPassword
V3Pf9SamlOkta = v3.OktaPassword
V3Pf9SamlOnelogin = v3.OneloginPassword

__all__ = ('V3Pf9ADFSPassword', 'V3Pf9SamlOkta', 'V3Pf9SamlOnelogin')



# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""OneLogin SAML authentication driver."""
import base64
import re
import urlparse
from pf9_saml_auth.v3 import base


class Password(base.BasePF9SAMLPlugin):
    """onelogin.Password class."""

    def __init__(self, onelogin_client_id, onelogin_client_secret, **kwargs):
        """Class constructor accepting following parameters.

        :param onelogin_client_id: OneLogin API Client ID
        :type onelogin_client_id: string
        :param onelogin_client_secret: OneLogin API Client Secret
        :type onelogin_client_secret: string

        Remaining arguments inherit from parent class.
        """
        self.client_id = onelogin_client_id
        self.client_secret = onelogin_client_secret
        self.session = None

        super(Password, self).__init__(**kwargs)

        self.api_endpoint = 'https://api.us.onelogin.com'

    def _app_info(self):
        """OneLogin-specific application info retrieved from the URL.

        :returns: tuple containing subdomain & APP ID
        :rtype: tuple
        """
        redirect_url = urlparse.urlparse(self._get_redirect_url())
        if re.search("onelogin", redirect_url.hostname):
            subdomain = re.match(
                r"^([a-z0-9\-]+).onelogin.com",
                redirect_url.hostname)
            app_id = re.match(
                r"^\/trust\/saml2\/http-redirect\/sso/(\d+)$",
                redirect_url.path)

            return (subdomain.group(1), app_id.group(1))

    def _authenticate(self, session):
        """Authenticate with identity provider.

        :param session:
        :type session: keystoneauth1.session.Session
        :returns: SAML response
        :rtype: str
        """
        if self.session is None:
            self.session = session

        access_token = self._get_oauth_token()
        subdomain, app_id = self._app_info()
        try:
            saml_response = self._get_saml_assertion(
                access_token,
                app_id,
                subdomain,
            )

            return saml_response
        except Exception as excp:
            raise excp

    def _get_oauth_token(self):
        """Get OAuth token to obtain API access."""
        url = self.api_endpoint + '/auth/oauth2/token'
        headers = {
            'Authorization': [
                'client_id:' + self.client_id,
                'client_secret:' + self.client_secret
            ]
        }

        # Flatten auth headers
        headers['Authorization'] = ','.join(headers['Authorization'])

        headers['Content-Type'] = 'application/json'
        payload = {'grant_type': 'client_credentials'}

        resp = self._post_call(self.session, url, headers, payload)
        for item in resp.json()['data']:
            if 'access_token' in item:
                access_token = item['access_token']
                return access_token

    def _get_saml_assertion(self, access_token, app_id, subdomain):
        """Authenticate with IdP & retrieve SAML assertion.

        :param access_token
        :type access_token: str
        :param app_id
        :type app_id: str
        :param subdomain
        :type subdomain: str
        """
        url = self.api_endpoint + '/api/1/saml_assertion'
        headers = {
            'Authorization': 'bearer:' + access_token,
            'Content-Type': 'application/json'
        }

        payload = {
            'username_or_email': self.username,
            'password': self.password,
            'app_id': app_id,
            'subdomain': subdomain
        }

        resp = self._post_call(self.session, url, headers, payload)
        saml_assertion = base64.b64decode(resp.json()['data'])

        return saml_assertion

    def mfa_supported(self):
        """
        Check if MFA is supported.

        :rtype: True if MFA supported, False otherwise
        """
        return False



# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os

from setuptools import setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fn:
        return fn.read()

setup(
    name='pf9-saml-auth',
    version='0.0.2',
    description='Platform9 SAML Authentication Library for OpenStack Identity',
    long_description=read('README.rst'),
    url='https://github.com/platform9/pf9-saml-auth',
    author='Blake Covarrubias',
    author_email='blake@platform9.com',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: OpenStack',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='openstack keystone saml',
    packages=[
        'pf9_saml_auth',
        'pf9_saml_auth.v3',
    ],
    install_requires=[
        'keystoneauth1',
        'lxml',
        'oktaauth',
        'python-keystoneclient',
    ],
    python_requires='>=2.7, !=3.*',
    zip_safe=False,
    entry_points={
        "keystoneauth1.plugin": [
            "v3pf9samladfs = pf9_saml_auth._loading:V3Pf9ADFSPassword",
            "v3pf9samlokta = pf9_saml_auth._loading:V3Pf9SamlOkta",
            "v3pf9samlonelogin = pf9_saml_auth._loading:V3Pf9SamlOnelogin",
        ]
    }
)



# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import oktaauth
from pf9_saml_auth.v3 import base
from keystoneauth1.extras._saml2 import V3ADFSPassword
from pf9_saml_auth.v3 import adfs
from pf9_saml_auth.v3 import okta
from pf9_saml_auth.v3 import onelogin

_ADFS_AVAILABLE = adfs is not None and V3ADFSPassword is not None
_OKTA_AVAILABLE = okta is not None and oktaauth is not None
_ONELOGIN_AVAILABLE = onelogin is not None

ADFSPassword = adfs.V3ADFSPassword
OktaPassword = okta.Password
OneloginPassword = onelogin.Password

__all__ = ('ADFSPassword', 'OktaPassword', 'OneloginPassword')



# Copyright 2018 Platform9 Systems, Inc.

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pf9_saml_auth
from keystoneauth1 import loading


class V3Pf9ADFSPassword(loading.BaseFederationLoader):

    @property
    def plugin_class(self):
        return pf9_saml_auth.V3Pf9ADFSPassword

    @property
    def available(self):
        return pf9_saml_auth._ADFS_AVAILABLE

    def get_options(self):
        options = super(V3Pf9ADFSPassword, self).get_options()

        options.extend([
            loading.Opt('service-provider-endpoint',
                        help="Service Provider's Endpoint"),
            loading.Opt('service-provider-entity-id',
                        help="Service Provider's SAML SP Entity ID"),
            loading.Opt('identity-provider-url',
                        help=('An Identity Provider URL, where the SAML2 '
                              'authentication request will be sent.')),
            loading.Opt('username', help='Username'),
            loading.Opt('password', secret=True, help='Password')
        ])

        return options


class V3Pf9SamlOkta(loading.BaseFederationLoader):

    @property
    def plugin_class(self):
        return pf9_saml_auth.V3Pf9SamlOkta

    @property
    def available(self):
        return pf9_saml_auth._OKTA_AVAILABLE

    def get_options(self):
        options = super(V3Pf9SamlOkta, self).get_options()

        options.extend([
            loading.Opt('username', help='Username'),
            loading.Opt('password', secret=True, help='Password')
        ])

        return options


class V3Pf9SamlOnelogin(loading.BaseFederationLoader):

    @property
    def plugin_class(self):
        return pf9_saml_auth.V3Pf9SamlOnelogin

    @property
    def available(self):
        return pf9_saml_auth._ONELOGIN_AVAILABLE

    def get_options(self):
        options = super(V3Pf9SamlOnelogin, self).get_options()

        options.extend([
            loading.Opt('username', help='Username'),
            loading.Opt('password', secret=True, help='Password'),
            loading.Opt(
                'onelogin-client-id',
                help='OneLogin API Client ID'
            ),
            loading.Opt(
                'onelogin-client-secret',
                secret=True,
                help='OneLogin API Client Secret'
            ),
        ])

        return options
